//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>

#if !defined(__AGCFILTERS__H__)
#define __AGCFILTERS__H__ 1
//#include <math.h>

#define FIRARRAYTYPE static const

#include "itqA.h"
#include "varray.h"
#include "vecmath.h"
#include "accuminfo.h"
#include "itq-qdefs.h"

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wconversion"

#pragma clang diagnostic pop

#define AGCFNOINLINE __attribute__((always_inline)) //__attribute__((noinline))
#define AGCFNOINLINE0 __attribute__((noinline))

template <typename VDEF>
inline varrayT<VDEF>
sratetmod(varrayT<VDEF> tval)
{
	return Vselect((tval == varrayT<VDEF>(1.0)), varrayT<VDEF>(1.0), tval);
}

template <typename T>
inline T
sratetmod(T tval)
{
	return (tval == T(1.0)) ? T(1.0) : (tval * T(aspec.SRATETMULT));
}

template <typename T, typename CT = double>
class hpfT
{
	T coeff;
	T zval;

public:
	hpfT(void) : coeff(), zval() {}

	hpfT(const hpfT<T, CT>& init) : coeff(init.coeff), zval(init.zval) {}

	hpfT(T c, T z = T{}): coeff(c), zval(z) {}
	
	hpfT(CT c, CT z = CT(0)) : hpfT(T(c), T(z)) {}

	hpfT<T, CT>&
	operator=(hpfT<T, CT> src)
	{
		coeff = src.coeff;
		zval = src.zval;
		return *this;
	}

	T
	next(T in)
	{
		T out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	T
	operator()(T in)
	{
		return next(in);
	}

	void
	setcoeff(T newcoeff)
	{
		coeff = newcoeff;
	}

	void
	setcoeff(CT newcoeff)
	{
		coeff = T(newcoeff);
	}
};

class hpfAs
{
	float coeff;
	Asample zval;

public:
	hpfAs(float c = 0.0f) :
		coeff(c), zval(0.0f, 0.0f)
	{}

	Asample
	next(const Asample& in)
	{
		Asample out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
};

class hpf4As
{
	float coeff[4];
	Asample zval[4];

public:

	hpf4As(float c0, float c1, float c2, float c3) :
		coeff(), zval()
	{
		for (int i = 0; i < 4; i++)
		{
			zval[i] = Asample{};
		}
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}

	hpf4As(float c0 = 0.0f) :
		hpf4As(c0, c0, c0, c0) {}

	hpf4As(varrayT<TYPEvd4x1> cvals) :
		hpf4As(float(cvals[0]), float(cvals[1]), float(cvals[2]), float(cvals[3])) {}

	hpf4As(varrayT<TYPEvf4x1> cvals) :
		hpf4As(cvals[0], cvals[1], cvals[2], cvals[3]) {}

	void
	setcoeffs(float c0, float c1, float c2, float c3)
	{
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}
		
	void
	next(const Asample* in, Asample* out)
	{
		for (int i = 0; i < 4; i++)
		{
			out[i] = in[i] - zval[i];
			zval[i] = zval[i] + out[i] * coeff[i];
		}
		return;
	}

	void
	operator()(const Asample* in, Asample* out)
	{
		next(in, out);
		return;
	}
};

template <int NFLT>
class XsfdnfAs GCQ
{
	static constexpr int REALNFLT = ((NFLT + 1) / 2) * 2;
	//static constexpr int REALNFLT = NFLT;
	int inptr;
	float rawbuf0[REALNFLT*2];
	float rawbuf1[REALNFLT*2];
	const float * filter;

	void
	newdata(const Asample& in)
	{
		float tmp = in.item0();
		rawbuf0[inptr] = tmp;
		rawbuf0[inptr+NFLT] = tmp;

		tmp = in.item1();
		rawbuf1[inptr] = tmp;
		rawbuf1[inptr+NFLT] = tmp;
		inptr += 1;
		if (inptr >= NFLT)
			inptr = 0;
	}

	Asample 
	basicfilt(void) const
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr0 = &rawbuf0[inptr];
		const float * rbptr1 = &rawbuf1[inptr];
		for (int i = 0; i < REALNFLT; i++)
		{
			float f = filter[i];
			out0 += rbptr0[i] * f;
			out1 += rbptr1[i] * f;
		}
		return Asample{out0, out1};
	}


public:
	XsfdnfAs(const float *filtera = nullptr) AGCFNOINLINE : inptr(0), rawbuf0(), rawbuf1(), filter(filtera) 
	{
		for (int i = 0; i < REALNFLT*2; i++)
		{
			rawbuf0[i] = 0.0f;
			rawbuf1[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample AGCFNOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}


};

template <int NFLT>
class sfdnfAs GCQ
{
	static constexpr int REALNFLT = ((NFLT + 1) / 2) * 2;
	float __attribute__((aligned(32))) rawbuf[REALNFLT*4];
	const float * __attribute__((aligned(32))) filter;

	Asample prev;
	int inptr;

	void
	newdata(const Asample& in)
	{
		float tmp0 = in.item0();
		float tmp1 = in.item1();
		rawbuf[inptr] = tmp0;
		rawbuf[inptr+1] = tmp1;
		rawbuf[inptr+NFLT*2] = tmp0;
		rawbuf[inptr+1+NFLT*2] = tmp1;

		inptr += 2;
		if (inptr >= NFLT*2)
			inptr = 0;
	}

	Asample 
	basicfilt(void)
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr = &rawbuf[inptr];
		for (int i = 0; i < REALNFLT; i += 1 )
		{
			float f = filter[i];
			int dindex = i * 2;
			out0 += rbptr[dindex] * f;
			out1 += rbptr[dindex+1] * f;
		}
#if 0
		if (NFLT & 1)
		{
			Asample rtval = prev;
			prev = Asample(out0, out1);
			return rtval;
		}
#endif
		return Asample{out0, out1};
	}


public:
	sfdnfAs(const float * __attribute__((aligned(32))) filtera = nullptr) AGCFNOINLINE : rawbuf(), filter(filtera), prev(), inptr(0)
	{
		for (int i = 0; i < REALNFLT*4; i++)
		{
			rawbuf[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample AGCFNOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV
	operator()(const PV& input)
	{
		return next(input);
	}


};

template <int NFLT, typename VDEF=TYPEvd4x1>
class sfdnfVa
{
	using VT=varrayT<VDEF>;
	using FT=typename VDEF::FT;
	VT prev;
	VT rawbuf[NFLT*2];
	const float * filter;
	int inptr;

	void
	newdata(VT in)
	{
		int nextinptr = inptr + 1;
		if (nextinptr >= NFLT)
			nextinptr = 0;

		rawbuf[inptr] = in;
		rawbuf[inptr+NFLT] = in;

		inptr = nextinptr;
	}

	VT 
	basicfilt(void) 
	{
		int i;
		VT fout = VT(FT(0.0));
		int rbbase = inptr;
		const VT * rbptr = (const VT *) &rawbuf[rbbase];
		for (i = 0; i < NFLT; i += 4)
		{
			fout += rbptr[i] * filter[i];
			fout += rbptr[i+1] * filter[i+1];
			fout += rbptr[i+2] * filter[i+2];
			fout += rbptr[i+3] * filter[i+3];
		}
		for (int j = i; j < NFLT; j++)
		{
			fout += rbptr[j] * filter[j];
		}
		VT rtval;
#if 0
		if (NFLT & 1)
		{
			rtval = prev;
			prev = fout;
		}
		else
#endif
		{
			rtval = fout;
		}
			
		return rtval;
	}


public:
	sfdnfVa(const float * const filtera = nullptr) AGCFNOINLINE : prev(FT(0)), inptr(0), rawbuf(), filter(filtera)
	{
		for (int i = 0; i < NFLT*2; i++)
			rawbuf[i] = VT(FT(0.0));
	}

	VT AGCFNOINLINE
	next(const VT& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	void
	setfilt(const float* const newfilt)
	{
		filter = newfilt;
	}
};

template <int NFLT, typename IOTYPE = TYPEvd4x1>
class sfdnfVd4x2X1
{
	using ft = sfdnfVa<NFLT,TYPEvf8x1>;

	using fltv = varrayT<TYPEvf8x1>;

	using iotypev = varrayT<IOTYPE>;
	using iotypeftype = typename IOTYPE::FT;
	

	ft filt;

public:
	sfdnfVd4x2X1(const float * const filtera = nullptr) : filt(filtera)
	{}


	void AGCFNOINLINE
	next(iotypev v0in, iotypev v1in, iotypev& v0out, iotypev& v1out)
	{
		fltv ftmpin {
			float(v0in[0]), float(v0in[1]), float(v0in[2]), float(v0in[3]),
			float(v1in[0]), float(v1in[1]), float(v1in[2]), float(v1in[3])
		};

		fltv ftmpout = filt(ftmpin);

		iotypev dtmpout0 {
			iotypeftype(ftmpout[0]), iotypeftype(ftmpout[1]), iotypeftype(ftmpout[2]), iotypeftype(ftmpout[3])
		};

		iotypev dtmpout1 {
			iotypeftype(ftmpout[4]), iotypeftype(ftmpout[5]), iotypeftype(ftmpout[6]), iotypeftype(ftmpout[7])
		};

		v0out = dtmpout0;
		v1out = dtmpout1;
	}

	
	template <typename PV>
	void
	operator()(const PV& input0, const PV& input1, PV& output0, PV& output1)
	{
		next(input0, input1, output0, output1);
	}


	void
	setfilt(const float* const newfilt)
	{
		filt.setfilt(newfilt);
	}
};
	

template <int MAXDLY, typename DT = float>
class sVdly GCQ
{
	int inptr;
	DT rawbuf[2*MAXDLY];

	void
	newdata(DT in)
	{
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdly() : inptr(0), rawbuf()
	{
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = DT{0};
		}
	}

	const DT&
	next(const DT& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	DT
	dly(int vdly) const
	{
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	DT
	next(DT in, int vdly)
	{
		newdata(in);
		return dly(vdly);
	}

	DT operator()(const DT& input)
	{
		return next(input);
	}

	DT operator()(const DT& input, int delay)
	{
		return next(input, delay);
	}

	const DT operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

	int maxdelay(void) const {
		return MAXDLY;
	}

};

template <int NDLY>
class sdlyAs GCQ
{
	int inptr;
	Asample rawbuf[(NDLY < 1)?1:NDLY];

	void
	newdata(const Asample& in)
	{
		rawbuf[inptr] = in;
		inptr += 1;
		if (inptr >= NDLY)
			inptr = 0;
	}

public:
	sdlyAs() : inptr(0), rawbuf()
	{
		for (int i = 0; i < NDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	Asample 
	next(const Asample& in)
	{
		if (NDLY < 1)
		{
			return in;
		}
		else
		{
			newdata(in);
			return rawbuf[inptr];
		}
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	int maxdelay(void) const {
		return NDLY;
	}

};

template <int MAXDLY>
class sVdlyAs GCQ
{
	int inptr;
	Asample rawbuf[2*MAXDLY];

	void
	newdata(const Asample& in)
	{
	//	::memcpy(&rawbuf[inptr], &in, sizeof(Asample));
	//	::memcpy(&rawbuf[inptr+MAXDLY], &in, sizeof(Asample));
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdlyAs() : inptr(0), rawbuf()
	{
#pragma clang loop unroll(disable)
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	const Asample&
	next(const Asample& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	const Asample& 
	next(const Asample& in, int vdly)
	{
		newdata(in);
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	const Asample&
	dly(int dlyamt) const
	{
		return rawbuf[inptr + (MAXDLY-1) - dlyamt];
	}

	const Asample&
	minusmax(int dlyamt) const
	{
		return dly(MAXDLY - dlyamt);
	}

	const Asample& operator()(const Asample& input)
	{
		return next(input);
	}

	const Asample& operator()(const Asample& input, int delay)
	{
		return next(input, delay);
	}

	const Asample& operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

	int maxdelay(void) const {
		return MAXDLY;
	}

};


#if 0
using FILT_DEFAULTVT = TYPEvd4x1;

static constexpr int BASEORDERMF = 1023;
static constexpr int BASEORDERMFL = 1279;
static constexpr int BASEORDERMFS = 767;

static constexpr int BASEORDERMF2 = 639;
static constexpr int BASEORDERMF2L = 767;
static constexpr int BASEORDERMF2S = 512;

static constexpr int BASEORDER = 511;
static constexpr int BASEORDERL = 639;
static constexpr int BASEORDERS = 383;
static constexpr int BASEORDERVS = 351;

static constexpr int BASEORDERSHORT = 255;
static constexpr int BASEORDERSHORTL = 351;
static constexpr int BASEORDERSHORTS = 191;
static constexpr int BASEORDERSHORTVS = 128;

static constexpr int BASEORDERMAX = BASEORDERMFL;

static constexpr int BASEORDERGAINVL = 511;
static constexpr int BASEORDERGAINL = 383;
static constexpr int BASEORDERGAIN = 319;
static constexpr int BASEORDERGAINS = 255;
static constexpr int BASEORDERGAINVS = 191;
static constexpr int BASEORDERGAINVVS = 127;

static constexpr int BASEORDERGAINMFL = 1151;
static constexpr int BASEORDERGAINMF = 767; 
static constexpr int BASEORDERGAINMFS = 639;
static constexpr int BASEORDERGAINMFVS = 511;
static constexpr int BASEORDERGAINMFVVS = 383;

static constexpr int BASEORDERGAINMF2L = 511;
static constexpr int BASEORDERGAINMF2 = 511;
static constexpr int BASEORDERGAINMF2S = 447;
static constexpr int BASEORDERGAINMF2VS = 383;
static constexpr int BASEORDERGAINMF2VVS = 255;


static constexpr float NORMTOPHF = 22000.0f;
static constexpr float MAXHF1 = 45000.0f;
//static constexpr float NORMTOPHF = 21750.0f;

// #define DELAYCALC	static constexpr int DELAY = (ORDER + 1) / 2;
#define DELAYCALC	static constexpr int DELAY = (ORDER) / 2;
//#define DELAYCALC2	static constexpr int DELAY = (((ORDER0 + 1) / 2) + ((ORDER1 + 1) / 2));
//#define DELAYCALC2	static constexpr int DELAY = (((ORDER0) / 2) + ((ORDER1) / 2));
#define DELAYCALC2	static constexpr int DELAY = ((ORDER0 + ORDER1 - 1) / 2);

#endif

#endif
//
// vim: set ts=4:
//


